#!/usr/bin/env python
# -*- coding: utf-8 -*-

# # 2A.eco - Rappel de ce que vous savez déjà mais avez peut-être oublié
# 
# [pandas](http://pandas.pydata.org/) et [numpy](http://www.numpy.org/) sont essentiels pour manipuler les données. C'est ce que rappelle ce notebook. Voir aussi [Essential Cheat Sheets for Machine Learning and Deep Learning Engineers](https://startupsventurecapital.com/essential-cheat-sheets-for-machine-learning-and-deep-learning-researchers-efb6a8ebd2e5).

# In[1]:


from jyquickhelper import add_notebook_menu
add_notebook_menu()


# ## Les quelques règles de Python
# 
# Python est un peu susceptible et protocolaire, il y a quelques règles à respecter : 
# 
# 1) L'indentation est primordiale : un code mal indenté ne fonctionne pas. 
# L'indentation indique à l'interpréteur où se trouvent les séparations entre des blocs d'instructions. Un peu comme des points dans un texte. 
# Si les lignes ne sont pas bien alignées, l'interpréteur ne sait plus à quel bloc associer la ligne.
# 
# 2) On commence à compter à 0. Ca peut paraitre bizarre mais c'est comme ça. Le premier élément d'une liste est le 0-ème. 
# 
# 3) Les marques de ponctuation sont importantes : 
# - Pour une liste : []
# - Pour un dictionnaire : {}
# - Pour un tuple : ()
# - Pour séparer des éléments : ,
# - Pour commenter un bout de code : #
# - Pour aller à la ligne dans un bloc d'instructions : \
# - Les majuscules et minuscules sont importantes
# - Par contre l'usage des ' ou des " est indifférente. Il faut juste avoir les mêmes début et fin.
# - Pour documenter une fonction ou une classe """ documentation """ 

# ## Les outputs de Python : l'opération, le print et le return
# 
# Quand Python réalise des opérations, il faut lui préciser ce qu'il doit en faire : 
# - est-ce qu'il doit juste faire l'opération,
# - afficher le résultat de l'opération, 
# - créer un objet avec le résultat de l'opération ?
# 
# 
# Remarque : dans l'environnement Notebook, le dernier élement d'une cellule est automatiquement affiché (print), qu'on lui demande ou non de le faire. Ce n'est pas le cas dans un éditeur classique comme Spyder.

# In[2]:


# on calcule : dans le cas d'une opération par exemple une somme
2+3 # Python calcule le résultat mais n'affiche rien dans la sortie

# le print : on affiche

print(2+3) # Python calcule et on lui demande juste de l'afficher
# le résultat est en dessous du code


# In[3]:


# le print dans une fonction 

def addition_v1(a,b) : 
    print(a+b)

resultat_print = addition_v1(2,0) 
print(type(resultat_print))

# dans la sortie on a l'affichage du résultat, car la sortie de la fonction est un print 
# en plus on lui demande quel est le type du résultat. Un print ne renvoie aucun type, ce n'est ni un numérique,
# ni une chaine de charactères, le résultat d'un print n'est pas un format utilisable


# Le résultat de l'addition est affiché
# 
# La fonction addition_v1 effectue un print
# 
# Par contre, l'objet crée n'a pas de type, il n'est pas un chiffre, ce n'est qu'un affichage.
# 
# Pour créer un objet avec le résultat de la fonction, il faut utiliser __return__

# In[4]:


# le return dans une fonction
def addition_v2(a,b) : 
    return a+b

resultat_return = addition_v2(2,5) # 
print(type(resultat_return))
## là on a bien un résultat qui est du type "entier"


# 
# Le résultat de addition_v2 n'est pas affiché comme dans addition_v1
# 
# Par contre, la fonction addition_v2 permet d'avoir un objet de type int, un entier donc.

# ## Type de base : variables, listes, dictionnaires ...
# 
# Pyhon permet de manipuler différents types de base
# 
# On distingue deux types de variables : les immuables qui ne peuvent être modifiés et les modifiables
# 
# ### Les variables - types immuables
# 
# Les variables immuables ne peuvent être modifiées
# 
# - None : ce type est une convention de programmation pour dire que la valeur n'est pas calculée
# - bool : un booléen
# - int : un entier
# - float : un réel
# - str : une chaine de caractères
# - tuple : un vecteur
# 
# 

# In[5]:


i = 3         # entier = type numérique (type int)
r = 3.3       # réel   = type numérique (type float)
s = "exemple" # chaîne de caractères = type str 
n = None      # None signifie que la variable existe mais qu'elle ne contient rien
              # elle est souvent utilisée pour signifier qu'il n'y a pas de résultat
a = (1,2)     # tuple

print(i,r,s,n,a)         


# Si on essaie de changer le premier élément de la chaine de caractères __s__ on va avoir un peu de mal. 
# 
# Par exemple si on voulait mettre une majuscule à "exemple", on aurait envie d'écrire que le premier élément de la chaine s est "E" majuscule
# 
# Mais Python ne va pas nous laisser faire, il nous dit que les objets "chaine de caractère" ne peuvent être modifiés

# In[6]:


s[0] = "E"  # déclenche une exception


# Tout ce qu'on peut faire avec une variable immutable, c'est le réaffecter à une autre valeur : il ne peut pas être modifié
# 
# Pour s'en convaincre, utilisons la fonction id() qui donne un identifiant à chaque objet.

# In[7]:


print(s)
id(s)


# In[8]:


s = "autre_mot"
id(s)


# On voit bien que s a changé d'identifiant : il peut avoir le même nom, ce n'est plus le même objet

# ### Les variables - types modifiable : listes et dictionnaires
# 
# Heureusement, il existe des variables modifiables comme les listes et les dictionnaires.  
# 
# #### Les listes - elles s''écrivent entre [ ]
# 
# 
# Les listes sont des élements très utiles, notamment quand vous souhaitez faire des boucles
# 
# Pour faire appel aux élements d'une liste, on donne leur position dans la liste : le 1er est le 0, le 2ème est le 1 ...

# In[9]:


ma_liste = [1,2,3,4]


print("La longueur de ma liste est de", len(ma_liste))

print("Le premier élément de ma liste est :", ma_liste[0])

print("Le dernier élément de ma liste est :", ma_liste[3])
print("Le dernier élément de ma liste est :", ma_liste[-1])


# #### Les dictionnaires - ils s'écrivent entre { }
# Un dictionnaire associe à une clé un autre élément, appelé une valeur : un chiffre, un nom, une liste, un autre dictionnaire etc.
# 
#   - __Format d'un dictionnaire : {Clé : valeur}__
#    

# #### Dictionnaire avec des valeurs int
# 
# On peut par exemple associer à un nom, un nombre

# In[10]:


mon_dictionnaire_notes = { 'Nicolas' : 18 , 'Pimprenelle' : 15} 
# un dictionnaire qui à chaque nom associe un nombre
# à Nicolas, on associe 18

print(mon_dictionnaire_notes) 


# #### Dictionnaire avec des valeurs qui sont des listes

# Pour chaque clé d'un dictionnaire, il ne faut pas forcément garder la même forme de valeur
# 
# Dans l'exemple, la valeur de la clé "Nicolas" est une liste, alors que celle de "Philou" est une liste de liste

# In[11]:


mon_dictionnaire_loisirs =  \
{ 'Nicolas' : ['Rugby','Pastis','Belote'] , 
  'Pimprenelle' : ['Gin Rami','Tisane','Tara Jarmon','Barcelone','Mickey Mouse'],
  'Philou' : [['Maths','Jeux'],['Guillaume','Jeanne','Thimothée','Adrien']]}


# Pour accéder à un élément du dictionnaire, on fait appel à la clé et non plus à la position, comme c'était le cas dans les listes

# In[12]:


print(mon_dictionnaire_loisirs['Nicolas']) # on affiche une liste


# In[13]:


print(mon_dictionnaire_loisirs['Philou']) # on affiche une liste de listes


# Si on ne veut avoir que la première liste des loisirs de Philou, on demande le premier élément de la liste

# In[14]:


print(mon_dictionnaire_loisirs['Philou'][0]) # on affiche alors juste la première liste


# On peut aussi avoir des valeurs qui sont des int et des listes

# In[15]:


mon_dictionnaire_patchwork_good = \
{ 'Nicolas' : ['Rugby','Pastis','Belote'] ,
  'Pimprenelle' : 18 }


# ## A retenir 
# 
#     
#    - L'indentation du code est importante (4 espaces et pas une tabulation)
#    - Une __liste__ est entre [] et on peut appeler les positions par leur place 
#    - Un __dictionnaire__, clé x valeur, s'écrit entre {} et on appelle un élément en fonction de la clé

# ## Questions pratiques : 
# 
#    - Quelle est la position de 7 dans la liste suivante

# In[16]:


liste_nombres = [1,2,7,5,3]


# - Combien de clés a ce dictionnaire ?     

# In[17]:


dictionnaire_evangile = {"Marc" : "Lion", "Matthieu" : ["Ange","Homme ailé"] , 
                         "Jean" : "Aigle" , "Luc" : "Taureau"}


# - Que faut-il écrire pour obtenir "Ange" en résultat à partir du dictionnaire_evangile ? 

# ## Objets : méthodes et attributs
# 
# Mainentant qu'on a vu quels objets existaient en Python, nous allons voir comment nous en servir.

# ### Un petit détour pour bien comprendre : Un objet, c'est quoi ? 
# 
# Un objet a deux choses : des attributs et des méthodes
# 
#    - Les attributs décrivent sa structure interne : sa taille, sa forme (dont on ne va pas parler ici)
#    - Les méthodes sont des  "actions" qui s'appliqueront à l'objet

# ### Premiers exemples de méthode
# 
# Avec les éléments définis dans la partie 1 (les listes, les dictionnaires) on peut faire appel à des méthodes qui sont directement liées à ces objets.
# 
# Les méthodes, c'est un peu les actions de Python. 

# #### Une méthode pour les listes 
# 
# Pour ajouter un item dans une liste : on va utiliser la méthode _.append()_

# In[18]:


ma_liste = ["Nicolas","Michel","Bernard"]

ma_liste.append("Philippe")

print(ma_liste)


# #### Une méthode pour les dictionnaires
# 
# Pour connaitre l'ensemble des clés d'un dictionnaire, on appelle la méthode _.keys()_

# In[19]:


mon_dictionnaire = {"Marc" : "Lion", "Matthieu" : ["Ange","Homme ailé"] , 
                          "Jean" : "Aigle" , "Luc" : "Taureau"}

print(mon_dictionnaire.keys())


# ### Connaitre les méthodes d'un objet
# 
# Pour savoir quelles sont les méthodes d'un objet vous pouvez : 
#   - taper help(mon_objet) ou mon_objet? dans la console iPython
#   - taper mon_objet. + touche tabulation dans la console iPython ou dans le notebook . iPython permet la complétion, c'est-à-dire que vous pouvez faire appaître la liste

# ## Les opérations et méthodes classiques des listes

# ### Créer une liste
# 
# Pour créer un objet de la classe list, il suffit de le déclarer. Ici on affecte à __x__ une liste

# In[20]:


x = [4, 5] # création d’une liste composée de deux entiers
x = ["un", 1, "deux", 2] # création d’une liste composée de 2 chaînes de caractères
# et de deux entiers, l’ordre d’écriture est important
x = [3] # création d’une liste d’un élément, sans la virgule,
x = [ ] # crée une liste vide
x = list () # crée une liste vide


# ### Un premier test sur les listes
# 
# Si on veut tester la présence d'un élément dans une liste, on l'écrit de la manière suivante : 

# In[21]:


# Exemple 

x = "Marcel"

l = ["Marcel","Edith","Maurice","Jean"]

print(x in l)

#vrai si x est un des éléments de l


# ### Pour concaténer deux listes : 
# On utilise le symbole +

# In[22]:


t = ["Antoine","David"]
print(l + t) #concaténation de l et t


# ### Pour trouver certains éléments d'une liste
# Pour chercher des élements dans une liste, on utilise la position dans la liste. 

# In[23]:


l[1] # donne l'élément qui est en 2ème position de la liste


# In[24]:


l[1:3] # donne les éléments de la 2ème position de la liste à la 4ème exclue


# ### Quelques fonctions des listes

# In[25]:


longueur = len(l) # nombre d’éléments de l

minimum = min(l) # plus petit élément de l, ici par ordre alphabétique

maximum = max(l) # plus grand élément de l, ici par ordre alphabétique

print(longueur,minimum,maximum)


# In[26]:


del l[0 : 2] # supprime les éléments entre la position 0 et 2 exclue
print(l)


# ## Les méthodes des listes
# 
# On les trouve dans l'aide de la liste. On distingue les méthodes et les méthodes spéciales : visuellement, les méthodes spéciales sont celles qui précédées et suivis de deux caractères de soulignement, les autres sont des méthodes classiques.
# 
# 

# In[27]:


help(l)


# ## A retenir et questions
# 
# A retenir : 
# 
# - Chaque objet Python a des attributs et des méthodes
# - Vous pouvez créer des classes avec des attributs et des méthodes 
# - Les méthodes des listes et des dictionnaires sont les plus utilisées : 
#     - list.count()
#     - list.sort()
#     - list.append()
#     - dict.keys()
#     - dict.items()
#     - dict.values()
#     
# ------

# Questions pratiques : 
# 
#  - Définir la liste allant de 1 à 10, puis effectuez les actions suivantes :
#  
#         – triez et affichez la liste 
#         – ajoutez l’élément 11 à la liste et affichez la liste 
#         – renversez et affichez la liste 
#         – affichez l’indice de l’élément 7 
#         – enlevez l’élément 9 et affichez la liste 
#         – affichez la sous-liste du 2e au 3e élément ;
#         – affichez la sous-liste du début au 2e élément ;
#         – affichez la sous-liste du 3e élément à la fin de la liste ;
# 

# - Construire le dictionnaire des 6 premiers mois de l'année avec comme valeurs le nombre de jours respectif. 
# 
#         - Renvoyer la liste des mois. 
#         - Renvoyer la liste des jours.
#         - Ajouez la clé du mois de Juillet ?
#  
# 

# ## Passer des listes, dictionnaires à pandas

# Supposons que la variable 'data' est un liste qui contient nos données. 
# 
# Une observation correspond à un dictionnaire qui contient le nom, le type, l'ambiance et la note d'un restaurant. 
# 
# Il est aisé de transformer cette liste en dataframe grâce à la fonction 'DataFrame'.

# In[28]:


import pandas 

data = [{"nom": "Little Pub", "type" : "Bar", "ambiance": 9, "note": 7},
     {"nom": "Le Corse", "type" : "Sandwicherie", "ambiance": 2, "note": 8},
     {"nom": "Café Caumartin", "type" : "Bar", "ambiance": 1}]

df = pandas.DataFrame(data)

print(data)
df


# In[29]: